# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains generic utility functions for the library."""
from __future__ import annotations

import os
from json import JSONEncoder
from typing import TYPE_CHECKING, Any, Union

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import NDArray

PathLike = Union[str, os.PathLike]


def _assert(
    condition: bool,
    error_message: str | None = None,
    exception_class: type[Exception] = Exception,
) -> None:
    """Check condition; if false, raise exception with the provided error message."""
    if not condition:
        raise exception_class(error_message)


def check_arg(condition: bool, message: str) -> None:
    """Check a condition over an argument (i.e. raises ValueError in case it is false).

    Args:
        condition: the condition to check
        message: the error message
    """
    _assert(condition, message, exception_class=ValueError)


def ensure_float(arg: Any) -> None:  # noqa: ANN401
    """Check that the argument is a float."""
    check_arg(isinstance(arg, (float, np.float64)), f"expected a float, got {arg}")


def positive_float(arg: float) -> float:
    """Check a float is positive."""
    ensure_float(arg)
    check_arg(arg >= 0, "provided float is not positive")
    return arg


def digitize_data(
    data: NDArray[np.float64],
    param_grid: list[NDArray[np.float64]],
) -> NDArray[np.float64]:
    """Return a discretized version of the input sorted_array.

    Args:
        data: the input array to be discretized
        param_grid: the discrete values allowed

    Returns:
        the discretized array
    """
    digitalized_data: NDArray[np.float64] = np.zeros(shape=data.shape)

    for i in range(data.shape[1]):
        digitalized_data[:, i] = get_closest(param_grid[i], data[:, i])

    return digitalized_data


def get_closest(sorted_array: NDArray, values: NDArray) -> NDArray:
    """Fast way to find the nearest element in 'sorted_array' for each element in 'values'.

    Solution taken from https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array/46184652#46184652.

    Args:
        sorted_array: a sorted array
        values: an array of arbitrary values

    Returns:
        for each element in the 'values' array the closest element in 'sorted_array' is returned
    """
    # get insert positions
    idxs = np.searchsorted(sorted_array, values, side="left")

    # find indexes where previous index is closer
    prev_idx_is_less = (idxs == len(sorted_array)) | (
        np.fabs(values - sorted_array[np.maximum(idxs - 1, 0)])
        < np.fabs(values - sorted_array[np.minimum(idxs, len(sorted_array) - 1)])
    )
    idxs[prev_idx_is_less] -= 1

    return sorted_array[idxs]


def is_symmetric(a: NDArray[np.float64]) -> bool:
    """Check if a matrix is symmetric.

    Args:
        a: the matrix

    Returns:
        True if the matrix is symmetric
    """
    return a.shape[0] == a.shape[1] and (a == a.T).all()


class NumpyArrayEncoder(JSONEncoder):
    """Custom JSONEncoder for Numpy arrays.

    Solution from https://pynative.com/python-serialize-numpy-ndarray-into-json/.
    """

    def default(self, o: Any) -> Any:  # noqa: ANN401
        """Implement custom JSON serialization for NumPy NDArrays."""
        if isinstance(o, np.ndarray):
            return o.tolist()
        return JSONEncoder.default(self, o)  # pragma: no cover



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the Gaussian process-based sampling."""

from __future__ import annotations

import warnings
from enum import Enum
from typing import TYPE_CHECKING, cast

import numpy as np
from scipy.special import erfc  # type: ignore[import]  # pylint: disable=no-name-in-module
from sklearn.gaussian_process import GaussianProcessRegressor, kernels  # type: ignore[import]

from black_it.samplers.surrogate import MLSurrogateSampler

if TYPE_CHECKING:
    from numpy.typing import NDArray


_BIG_DATASET_SIZE_WARNING_THRESHOLD = 500
_SMALL_VARIANCE_VALUES = 1e-5


class _AcquisitionTypes(Enum):
    """Enumeration of allowed acquisition types."""

    MEAN = "mean"
    EI = "expected_improvement"

    def __str__(self) -> str:
        """Get the string representation."""
        return self.value


class GaussianProcessSampler(MLSurrogateSampler):
    """This class implements the Gaussian process-based sampler.

    In particular, the sampling is based on a Gaussian Process interpolation of the loss function.

    Note: this class is a wrapper of the GaussianProcessRegressor model of the scikit-learn package.
    """

    def __init__(  # noqa: PLR0913
        self,
        batch_size: int,
        random_state: int | None = None,
        max_deduplication_passes: int = 5,
        candidate_pool_size: int | None = None,
        optimize_restarts: int = 5,
        acquisition: str = "expected_improvement",
        jitter: float = 0.1,
    ) -> None:
        """Initialize the sampler.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the random state of the sampler, fixing this number the sampler behaves deterministically
            max_deduplication_passes: the maximum number of deduplication passes that are made
            candidate_pool_size: number of randomly sampled points on which the random forest predictions are evaluated
            optimize_restarts: number of independent random trials of the optimization of the GP hyperparameters
            acquisition: type of acquisition function, it can be 'expected_improvement' of simply 'mean'
            jitter: positive value to make the "expected_improvement" acquisition more explorative.
        """
        self._validate_acquisition(acquisition)

        super().__init__(
            batch_size,
            random_state,
            max_deduplication_passes,
            candidate_pool_size,
        )
        self.optimize_restarts = optimize_restarts
        self.acquisition = acquisition
        self.jitter = jitter
        self._gpmodel: GaussianProcessRegressor | None = None
        self._fmin: np.double | float | None = None

    @staticmethod
    def _validate_acquisition(acquisition: str) -> None:
        """Check that the required acquisition is among the supported ones.

        Args:
            acquisition: the acquisition provided as input of the constructor.

        Raises:
            ValueError: if the provided acquisition type is not among the allowed ones.
        """
        try:
            _AcquisitionTypes(acquisition)
        except ValueError as e:
            msg = (
                "expected one of the following acquisition types: "
                f"[{' '.join(map(str, _AcquisitionTypes))}], got {acquisition}"
            )
            raise ValueError(
                msg,
            ) from e

    def fit(self, X: NDArray[np.float64], y: NDArray[np.float64]) -> None:  # noqa: N803
        """Fit a gaussian process surrogate model."""
        y = np.atleast_2d(y).T

        if X.shape[0] > _BIG_DATASET_SIZE_WARNING_THRESHOLD:
            warnings.warn(  # noqa: B028
                "Standard GP evaluations can be expensive for large datasets, consider implementing a sparse GP",
                RuntimeWarning,
            )

        # initialize GP class from scikit-learn with a Matern kernel
        kernel = kernels.Matern(length_scale=1, length_scale_bounds=(1e-5, 1e5), nu=2.5)

        noise_var = y.var() * 0.01

        self._gpmodel = GaussianProcessRegressor(
            kernel=kernel,
            alpha=noise_var,
            n_restarts_optimizer=self.optimize_restarts,
            optimizer="fmin_l_bfgs_b",
            random_state=self._get_random_seed(),
        )

        self._gpmodel.fit(X, y)

        # store minimum
        if self.acquisition == "expected_improvement":
            m, _ = self._predict_mean_std(X)
            self._fmin = np.min(m)

    def predict(self, X: NDArray[np.float64]) -> NDArray[np.float64]:  # noqa: N803
        """Predict using a gaussian process surrogate model."""
        # predict mean or expected improvement on the full sample set
        if self.acquisition == _AcquisitionTypes.EI.value:
            # minus sign needed for subsequent sorting
            candidates_score = -self._predict_EI(X, self.jitter)
        else:  # acquisition is "mean"
            candidates_score = self._predict_mean_std(X)[0]

        return candidates_score

    def _predict_mean_std(
        self,
        X: NDArray[np.float64],  # noqa: N803
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Predict mean and standard deviation of a fitted GP.

        Args:
            X: the points on which the predictions should be performed

        Returns:
            The pair (mean, std).
        """
        gpmodel = cast(GaussianProcessRegressor, self._gpmodel)
        X = X[None, :] if X.ndim == 1 else X  # noqa: N806
        m, s = gpmodel.predict(X, return_std=True, return_cov=False)
        s = np.clip(s, 1e-5, np.inf)
        return m, s

    def _predict_EI(  # noqa: N802
        self,
        X: NDArray[np.float64],  # noqa: N803
        jitter: float = 0.1,
    ) -> NDArray[np.float64]:
        """Compute the Expected Improvement per unit of cost.

        Args:
            X:  the points on which the predictions should be performed
            jitter: positive value to make the acquisition more explorative.

        Returns:
            the expected improvement.
        """
        m, s = self._predict_mean_std(X)

        fmin = cast(float, self._fmin)

        phi, Phi, u = self.get_quantiles(jitter, fmin, m, s)  # noqa: N806

        return s * (u * Phi + phi)

    @staticmethod
    def get_quantiles(
        acquisition_par: float,
        fmin: float,
        m: NDArray[np.float64],
        s: NDArray[np.float64],
    ) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
        """Quantiles of the Gaussian distribution useful to determine the acquisition function values.

        Args:
            acquisition_par: parameter of the acquisition function
            fmin: current minimum.
            m: vector of means.
            s: vector of standard deviations.

        Returns:
            the quantiles.
        """
        # remove values of variance that are too small
        s[s < _SMALL_VARIANCE_VALUES] = _SMALL_VARIANCE_VALUES

        u: NDArray[np.float64] = (fmin - m - acquisition_par) / s
        phi: NDArray[np.float64] = np.exp(-0.5 * u**2) / np.sqrt(2 * np.pi)
        Phi: NDArray[np.float64] = 0.5 * erfc(-u / np.sqrt(2))  # noqa: N806

        return phi, Phi, u



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains serialization and deserialization of calibration state with Pandas."""
from __future__ import annotations

import json
import pickle  # nosec B403
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd  # type: ignore[import]
import tables  # type: ignore[import]

from black_it.utils.base import NumpyArrayEncoder, PathLike

if TYPE_CHECKING:
    from collections.abc import Mapping

    from numpy.typing import NDArray

    from black_it.loss_functions.base import BaseLoss
    from black_it.schedulers.base import BaseScheduler


def load_calibrator_state(checkpoint_path: PathLike, _code_state_version: int) -> tuple:
    """Load calibrator data from a given folder.

    Args:
        checkpoint_path: the folder where the data are stored
        _code_state_version: the serialization version (unused)

    Returns:
        all the data needed to reconstruct the calibrator state
    """
    checkpoint_path = Path(checkpoint_path)
    with (checkpoint_path / "calibration_params.json").open() as f:
        cp = json.load(f)

    cr = pd.read_csv(checkpoint_path / "calibration_results.csv")

    params_samp_list = [
        cr[f"params_samp_{i}"] for i in range(len(cp["parameters_precision"]))
    ]

    params_samp = np.vstack(params_samp_list).T

    with (checkpoint_path / "scheduler_pickled.pickle").open("rb") as fb:
        scheduler = pickle.load(fb)  # nosec B301

    with (checkpoint_path / "loss_function_pickled.pickle").open("rb") as fb:
        loss_function = pickle.load(fb)  # nosec B301

    series_filename = checkpoint_path / "series_samp.h5"
    series_file = tables.open_file(str(series_filename), mode="r")
    series_samp = series_file.root.data[:]
    series_file.close()

    return (
        # initialization parameters
        cp["parameters_bounds"],
        cp["parameters_precision"],
        np.asarray(cp["real_data"]),
        cp["ensemble_size"],
        cp["N"],
        cp["D"],
        cp["convergence_precision"],
        cp["verbose"],
        cp["saving_file"],
        cp["initial_random_seed"],
        cp["random_generator_state"],
        cp["model_name"],
        scheduler,
        loss_function,
        cp["current_batch_index"],
        cp["n_sampled_params"],
        cp["n_jobs"],
        # calibration results
        params_samp,
        cr["losses_samp"].to_numpy(),
        series_samp,
        cr["batch_num_samp"].to_numpy(),
        cr["method_samp"].to_numpy(),
    )


def save_calibrator_state(  # noqa: PLR0913
    checkpoint_path: PathLike,
    parameters_bounds: NDArray[np.float64],
    parameters_precision: NDArray[np.float64],
    real_data: NDArray[np.float64],
    ensemble_size: int,
    N: int,  # noqa: N803
    D: int,  # noqa: N803
    convergence_precision: float | None,
    verbose: bool,
    saving_file: str | None,
    initial_random_seed: int | None,
    random_generator_state: Mapping,
    model_name: str,
    scheduler: BaseScheduler,
    loss_function: BaseLoss,
    current_batch_index: int,
    n_sampled_params: int,
    n_jobs: int,
    params_samp: NDArray[np.float64],
    losses_samp: NDArray[np.float64],
    series_samp: NDArray[np.float64],
    batch_num_samp: NDArray[np.int64],
    method_samp: NDArray[np.int64],
) -> None:
    """Store the state of the calibrator in a given folder.

    Args:
        checkpoint_path: path to the checkpoint
        parameters_bounds: the parameters bounds
        parameters_precision: the parameters precision
        real_data: the real data
        ensemble_size: the ensemble size
        N: the number of samples
        D: the number of dimensions
        convergence_precision: the convergence precision
        verbose: the verbosity mode
        saving_file: the saving file
        initial_random_seed: the initial seed of the calibrator
        random_generator_state: the internal random state of the calibrator
        model_name: the model name
        scheduler: the scheduler to use in the calibration
        loss_function: the loss function
        current_batch_index: the current batch index
        n_sampled_params: the number of sampled params
        n_jobs: the number of jobs
        params_samp: the sampled parameters
        losses_samp: the sampled losses
        series_samp: the sampled series
        batch_num_samp: the sampling batch number
        method_samp: the sampling method
    """
    checkpoint_path = Path(checkpoint_path)
    # create directory if needed
    if not checkpoint_path.exists():
        checkpoint_path.mkdir(parents=True)

    calibration_params = {
        "parameters_bounds": parameters_bounds,
        "parameters_precision": parameters_precision,
        "real_data": real_data,
        "ensemble_size": ensemble_size,
        "N": N,
        "D": D,
        "convergence_precision": convergence_precision,
        "verbose": verbose,
        "saving_file": saving_file,
        "initial_random_seed": initial_random_seed,
        "random_generator_state": random_generator_state,
        "model_name": model_name,
        "current_batch_index": current_batch_index,
        "n_sampled_params": n_sampled_params,
        "n_jobs": n_jobs,
    }
    # save calibration parameters in a json dictionary
    with (checkpoint_path / "calibration_params.json").open("w") as f:
        json.dump(calibration_params, f, cls=NumpyArrayEncoder)

    # save instantiated scheduler and loss functions
    with (checkpoint_path / "scheduler_pickled.pickle").open("wb") as fb:
        pickle.dump(scheduler, fb)

    with (checkpoint_path / "loss_function_pickled.pickle").open("wb") as fb:
        pickle.dump(loss_function, fb)

    # save calibration results into pandas dataframe
    calibration_results = {
        "losses_samp": losses_samp.tolist(),
        "batch_num_samp": batch_num_samp.tolist(),
        "method_samp": method_samp.tolist(),
    }

    for d in range(params_samp.shape[1]):
        param_key_d = f"params_samp_{d}"
        calibration_results[param_key_d] = params_samp[:, d]

    df_calibration_results = pd.DataFrame.from_dict(calibration_results)
    df_calibration_results.to_csv(checkpoint_path / "calibration_results.csv")

    series_filename = "series_samp.h5"
    series_filepath = checkpoint_path / series_filename
    if series_filepath.exists():
        series_file = tables.open_file(str(series_filepath), mode="a")
        previous_shape = series_file.root.data.shape
        nb_rows = previous_shape[0]
        to_append = series_samp[nb_rows:]
        series_file.root.data.append(to_append)
        series_file.close()
        return

    series_file = tables.open_file(str(series_filepath), mode="w")
    atom = tables.Float64Atom()
    array_c = series_file.create_earray(
        series_file.root,
        "data",
        atom,
        (0, *series_samp.shape[1:]),
    )
    array_c.append(series_samp)
    series_file.close()
    return



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module defines the calibration gym environment for MAB algorithms."""

from numpy._typing import NDArray

from black_it.schedulers.rl.envs.base import CalibrationEnv


class MABCalibrationEnv(CalibrationEnv[int]):
    """A calibration environment for MAB algorithms."""

    def get_next_observation(self) -> int:  # noqa: PLR6301
        """Get the next observation."""
        return 0

    def reset_state(self) -> int:  # noqa: PLR6301
        """Get the initial state."""
        return 0

    def get_reward(
        self,
        best_param: NDArray,  # noqa: ARG002
        best_loss: float,
    ) -> float:
        """Get the reward."""
        if self._curr_best_loss is None:
            msg = "cannot get reward, curr_best_loss should be already set"
            raise ValueError(msg)
        reward = 0.0
        if best_loss < self._curr_best_loss:
            reward = (self._curr_best_loss - best_loss) / self._curr_best_loss
            self._curr_best_loss = best_loss
        return reward



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Base module for RL agents."""

from abc import ABC, abstractmethod
from typing import Generic, SupportsFloat

from gymnasium.core import ActType, ObsType

from black_it.utils.seedable import BaseSeedable


class Agent(Generic[ObsType, ActType], BaseSeedable, ABC):
    """Interface for RL agents."""

    @abstractmethod
    def learn(
        self,
        state: ObsType,
        action: ActType,
        reward: SupportsFloat,
        next_state: ObsType,
    ) -> None:
        """Learn from an agent-environment interaction timestep.

        Args:
            state: the observation from where the action was taken
            action: the chosen action
            reward: the observed reward
            next_state: the next observation
        """
        raise NotImplementedError

    @abstractmethod
    def policy(self, state: ObsType) -> ActType:
        """Get the next action.

        Args:
            state: the state from where to take the action

        Returns:
            the chosen action.
        """
        raise NotImplementedError



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module includes the implementation of a non-stationary epsilon-greedy agent."""
from __future__ import annotations

from typing import SupportsFloat, cast

import numpy as np

from black_it.schedulers.rl.agents.base import Agent


class MABEpsilonGreedy(Agent[int, int]):
    """Implementation of a MAB eps-greedy algorithm."""

    def __init__(
        self,
        n_actions: int,
        alpha: float,
        eps: float,
        initial_values: float = 0.0,
        random_state: int | None = None,
    ) -> None:
        """Initialize the agent object.

        Args:
            n_actions: the number of actions
            alpha: the learning rate
            eps: the epsilon parameter
            initial_values: the initial value for the Q-function
            random_state: the random state
        """
        super().__init__(random_state=random_state)
        self.n_actions = n_actions
        self.actions_count = [0] * self.n_actions

        self.Q = [initial_values] * self.n_actions
        self.alpha = alpha
        self.eps = eps
        self.initial_values = initial_values

    def get_step_size(self, action: int) -> float:
        """Get the step size."""
        return 1 / self.actions_count[action] if self.alpha == -1 else self.alpha

    def learn(
        self,
        state: int,  # noqa: ARG002
        action: int,
        reward: SupportsFloat,
        next_state: int,  # noqa: ARG002
    ) -> None:
        """Learn from an agent-environment interaction timestep."""
        self.actions_count[action] += 1

        step_size = self.get_step_size(action)

        # do the learning
        self.Q[action] += step_size * (cast(float, reward) - self.Q[action])

    def policy(self, _obs: int) -> int:
        """Get the action for this observation."""
        best_action = np.argmax(self.Q)

        random_e = self.random_generator.random()
        if not random_e < self.eps:
            action = best_action
        else:
            options = np.arange(self.n_actions)
            action = self.random_generator.choice(options, 1)[0]

        return int(action)

    def reset(self) -> None:
        """Reset the agent."""
        self.Q = [0.0] * self.n_actions
        self.actions_count = [0] * self.n_actions



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation the CORS sampler.

Regis, Rommel G., and Christine A. Shoemaker. "Constrained global optimization of expensive black box functions
using radial basis functions." Journal of Global optimization 31.1 (2005): 153-171.

"""
from __future__ import annotations

from math import factorial
from typing import TYPE_CHECKING, Callable, cast

import numpy as np
import scipy.optimize as op  # type: ignore[import]
from numpy.linalg import LinAlgError
from numpy.typing import NDArray

from black_it.samplers.base import BaseSampler
from black_it.utils.base import digitize_data, positive_float

if TYPE_CHECKING:
    from black_it.search_space import SearchSpace


def volume_d_dimensional_ball_radius_1(dims: int) -> float:
    """Compute the volume of a d-dimensional ball with radius 1."""
    if dims % 2 == 0:
        return np.pi ** (dims / 2) / factorial(int(dims / 2))
    return (
        2
        * (4 * np.pi) ** ((dims - 1) / 2)
        * factorial(int((dims - 1) / 2))
        / factorial(dims)
    )


def cubetobox(
    X: NDArray[np.float64],  # noqa: N803
    space_bounds: NDArray,
) -> NDArray[np.float64]:
    """Go from normalized values (unit cube) to absolute values (box)."""
    return space_bounds[0] + X * (space_bounds[1] - space_bounds[0])


def boxtocube(
    X: NDArray[np.float64],  # noqa: N803
    space_bounds: NDArray,
) -> NDArray[np.float64]:
    """Go from absolute values (box) to normalized values (unit cube)."""
    return (X - space_bounds[0]) / (space_bounds[1] - space_bounds[0])


def rbf(points: NDArray[np.float64], losses: NDArray[np.float64]) -> Callable:
    """Build RBF-fit for given points (see Holmstrom, 2008 for details).

    Implementation mostly taken from:
    https://github.com/paulknysh/blackbox/blob/cd8baa0cc344f36b3fddf910ae8037f62009619c/black_box/blackbox.py#L148-L195

    Args:
        points: Array of multi-d points with corresponding values [[x1, x2, .., xd, val], ...].
        losses: array of losses for each point.

    Returns:
        callable function that returns the value of the RBF-fit at a given point.
    """
    n = len(points)
    d = len(points[0])

    def phi(r: float) -> float:
        """Compute phi."""
        return r * r * r

    phis = [
        [phi(np.linalg.norm(np.subtract(points[i], points[j]))) for j in range(n)]  # type: ignore[arg-type]
        for i in range(n)
    ]

    P = np.ones((n, d + 1))  # noqa: N806
    P[:, 0:-1] = points

    F = losses  # noqa: N806

    M = np.zeros((n + d + 1, n + d + 1))  # noqa: N806
    M[0:n, 0:n] = phis
    M[0:n, n : n + d + 1] = P
    M[n : n + d + 1, 0:n] = np.transpose(P)

    v = np.zeros(n + d + 1)
    v[0:n] = F

    try:
        sol = np.linalg.solve(M, v)
    except LinAlgError:
        # might help with singular matrices
        print(
            "Singular matrix occurred during RBF-fit construction. RBF-fit might be inaccurate!",
        )
        sol = np.linalg.lstsq(M, v)[0]

    lam, b, a = sol[0:n], sol[n : n + d], sol[n + d]

    def fit(x: float) -> float:
        return (
            sum(
                lam[i] * phi(np.linalg.norm(np.subtract(x, points[i])))  # type: ignore[arg-type]
                for i in range(n)
            )
            + np.dot(b, x)
            + a
        )

    return fit


class CORSSampler(BaseSampler):
    """Implement the modified CORS sampler."""

    def __init__(
        self,
        batch_size: int,
        max_samples: int,
        rho0: float = 0.5,
        p: float = 1.0,
        random_state: int | None = None,
        verbose: bool = False,
    ) -> None:
        """Initialize the CORS sampler.

        Args:
            batch_size: the batch size
            max_samples: the maximum number samples used to compute the density decay
            rho0: initial "balls density"
            p: rate of "balls density" decay (p=1 - linear, p>1 - faster, 0<p<1 - slower)
            random_state: the random state of the sampler
            verbose: activate verbose mode
        """
        super().__init__(
            batch_size,
            random_state=random_state,
            max_deduplication_passes=0,
        )
        self._max_samples = max_samples
        self._rho0 = positive_float(rho0)
        self._p = positive_float(p)
        self._verbose = verbose

        self._batch_id = 0

    @property
    def rho0(self) -> float:
        """Get the rho0 parameter."""
        return self._rho0

    @property
    def p(self) -> float:
        """Get the p parameter."""
        return self._p

    def sample_batch(
        self,
        batch_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Sample a batch of points."""
        nb_seed_points = len(existing_points)
        fmax = np.max(np.abs(existing_losses))
        v1 = volume_d_dimensional_ball_radius_1(search_space.dims)

        # the first time sample batch is called. Normalize seed_points
        current_points = boxtocube(existing_points, search_space.parameters_bounds)
        current_losses = existing_losses / fmax

        # fit RBF to all data
        fit = rbf(cast(NDArray[np.float64], current_points), current_losses)

        # enlarge array for new points
        current_points = np.append(
            cast(NDArray[np.float64], current_points),
            np.zeros((batch_size, search_space.dims)),
            axis=0,
        )

        for j in range(batch_size):
            r = (
                (
                    self.rho0
                    * (
                        (self._max_samples - 1.0 - (self._batch_id * batch_size + j))
                        / (self._max_samples - 1.0)
                    )
                    ** self.p
                )
                / (v1 * (nb_seed_points + self._batch_id * batch_size + j))
            ) ** (1.0 / search_space.dims)

            if self._verbose:
                print(f"Using radius {r}")
            cons = [
                {
                    "type": "ineq",
                    "fun": lambda x, localk=k: np.linalg.norm(
                        np.subtract(x, current_points[localk]),
                    )
                    - r,  # noqa: B023
                }
                for k in range(nb_seed_points + j)
            ]
            while True:
                minfit = op.minimize(
                    fit,
                    self.random_generator.random(search_space.dims),
                    method="SLSQP",
                    bounds=[[0.0, 1.0]] * search_space.dims,
                    constraints=cons,
                )
                if not np.isnan(minfit.x)[0]:
                    break

            current_points[nb_seed_points + j] = np.copy(minfit.x)

        self._batch_id += 1
        new_cube_batch = current_points[-batch_size:]
        new_box_batch = cubetobox(new_cube_batch, search_space.parameters_bounds)
        return digitize_data(new_box_batch, search_space.param_grid)



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module implements the 'RoundRobinScheduler' scheduler."""

import numpy as np
from numpy._typing import NDArray

from black_it.samplers.base import BaseSampler
from black_it.schedulers.base import BaseScheduler


class RoundRobinScheduler(BaseScheduler):
    """This class implement a simple round-robin sampler scheduler.

    The round-robin scheduler takes in input a list of samplers [S_0, S_1, ..., S_{n-1}],
      and, at batch i, it proposes the (i % n)-th sampler.
    """

    def __init__(  # type: ignore[no-untyped-def]
        self,
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> None:
        """Initialize the round-robin scheduler."""
        super().__init__(*args, **kwargs)

        self._batch_id = 0

    def get_next_sampler(self) -> BaseSampler:
        """Get the next sampler."""
        return self.samplers[self._batch_id % len(self.samplers)]

    def update(
        self,
        batch_id: int,  # noqa: ARG002
        new_params: NDArray[np.float64],  # noqa: ARG002
        new_losses: NDArray[np.float64],  # noqa: ARG002
        new_simulated_data: NDArray[np.float64],  # noqa: ARG002
    ) -> None:
        """Update the state of the scheduler after each batch."""
        self._batch_id += 1



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module implements the 'RLScheduler' scheduler."""
from __future__ import annotations

import threading
from typing import TYPE_CHECKING, cast

import numpy as np

from black_it.samplers.base import BaseSampler
from black_it.samplers.halton import HaltonSampler
from black_it.schedulers.base import BaseScheduler

if TYPE_CHECKING:
    from collections.abc import Sequence
    from queue import Queue

    from numpy._typing import NDArray

    from black_it.schedulers.rl.agents.base import Agent
    from black_it.schedulers.rl.envs.base import CalibrationEnv


class RLScheduler(BaseScheduler):
    """This class implement a RL-based scheduler.

    It is agnostic wrt the RL algorithm being used.
    """

    def __init__(
        self,
        samplers: Sequence[BaseSampler],
        agent: Agent,
        env: CalibrationEnv,
        random_state: int | None = None,
    ) -> None:
        """Initialize the scheduler."""
        self._original_samplers = samplers
        new_samplers, self._halton_sampler_id = self._add_or_get_bootstrap_sampler(
            samplers,
        )

        self._agent = agent
        self._env = env

        super().__init__(new_samplers, random_state)

        self._in_queue: Queue = self._env._out_queue  # noqa: SLF001
        self._out_queue: Queue = self._env._in_queue  # noqa: SLF001

        self._best_param: float | None = None
        self._best_loss: float | None = None

        self._agent_thread: threading.Thread | None = None
        self._stopped: bool = True

    def _set_random_state(self, random_state: int | None) -> None:
        """Set the random state (private use)."""
        super()._set_random_state(random_state)
        for sampler in self.samplers:
            sampler.random_state = self._get_random_seed()
        self._agent.random_state = self._get_random_seed()
        self._env.reset(seed=self._get_random_seed())

    @classmethod
    def _add_or_get_bootstrap_sampler(
        cls,
        samplers: Sequence[BaseSampler],
    ) -> tuple[Sequence[BaseSampler], int]:
        """Add or retrieve a sampler for bootstrapping.

        Many samplers do require some "bootstrapping" of the calibration process, i.e. a set of parameters
          whose loss has been already evaluated, e.g. samplers based on ML surrogates or on evolutionary approaches.
          Therefore, this scheduler must guarantee that the first proposed sampler is one that does not need previous
          model evaluations in input. One of such samplers is the Halton sampler

        Therefore, this function checks that the HaltonSampler is present in the set of samplers. If so, it returns
          the same set of samplers, and the index corresponding to that sampler in the sequence. Otherwise, a new
          instance of HaltonSampler is added to the list as first element.

        Args:
            samplers: the list of available samplers

        Returns:
            The pair (new_samplers, halton_sampler_id).
        """
        sampler_types = {type(s): i for i, s in enumerate(samplers)}
        if HaltonSampler in sampler_types:
            return samplers, sampler_types[HaltonSampler]

        new_sampler = HaltonSampler(batch_size=1)
        return tuple(list(samplers) + cast(list[BaseSampler], [new_sampler])), len(
            samplers,
        )

    def _train(self) -> None:
        """Run the training loop."""
        state = self._env.reset()
        while not self._stopped:
            # Get the action chosen by the agent
            action = self._agent.policy(state)
            # Interact with the environment
            next_state, reward, _, _, _ = self._env.step(action)
            # Learn from interaction
            self._agent.learn(state, action, reward, next_state)
            state = next_state

    def start_session(self) -> None:
        """Set up the scheduler for a new session."""
        if not self._stopped:
            msg = "cannot start session: the session has already started"
            raise ValueError(msg)
        self._stopped = False
        self._agent_thread = threading.Thread(target=self._train)
        self._agent_thread.start()

    def get_next_sampler(self) -> BaseSampler:
        """Get the next sampler."""
        if self._best_loss is None:
            # first call, return halton sampler
            return self.samplers[self._halton_sampler_id]
        chosen_sampler_id = self._in_queue.get()
        return self.samplers[chosen_sampler_id]

    def update(
        self,
        batch_id: int,  # noqa: ARG002
        new_params: NDArray[np.float64],
        new_losses: NDArray[np.float64],
        new_simulated_data: NDArray[np.float64],  # noqa: ARG002
    ) -> None:
        """Update the RL scheduler."""
        best_new_loss = float(np.min(new_losses))
        if self._best_loss is None:
            self._best_loss = best_new_loss
            self._best_param = new_params[np.argmin(new_losses)]
            self._env._curr_best_loss = best_new_loss  # noqa: SLF001
            return
        if best_new_loss < cast(float, self._best_loss):
            self._best_loss = best_new_loss
            self._best_param = new_params[np.argmin(new_losses)]

        self._out_queue.put((self._best_param, self._best_loss))

    def end_session(self) -> None:
        """Tear down the scheduler at the end of the session."""
        if self._stopped:
            msg = "cannot start session: the session has not started yet"
            raise ValueError(msg)
        self._stopped = True
        self._out_queue.put(None)
        cast(threading.Thread, self._agent_thread).join()



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains utility functions to deal with time series."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import scipy.sparse as sps  # type: ignore[import]
import statsmodels.api as sm  # type: ignore[import]
from scipy.stats import kurtosis, skew  # type: ignore[import]

if TYPE_CHECKING:
    from numpy.typing import NDArray


def get_mom_ts(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """Compute specific moments from a time series.

    Args:
        time_series: the time series

    Returns:
        the moments
    """
    moments = np.array([get_mom_ts_1d(ts_d) for ts_d in time_series.T])
    return moments.T


def get_mom_ts_1d(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """Compute specific moments from a time series.

    Args:
        time_series: the time series

    Returns:
        the moments from a time series
    """
    # array of all moments (to be filled)
    avg_vec_mom = np.zeros(18)

    # first 4 moments and auto-correlations of the time series
    avg_vec_mom[0] = np.mean(time_series)
    avg_vec_mom[1] = np.std(time_series)
    s, k = skew(time_series), kurtosis(time_series)
    avg_vec_mom[2] = np.sign(s) * np.power(abs(s), 1.0 / 3.0)
    avg_vec_mom[3] = np.sign(k) * np.power(abs(k), 1.0 / 4.0)

    ts_acf = sm.tsa.acf(time_series, nlags=5, fft=False)
    avg_vec_mom[4] = ts_acf[1]
    avg_vec_mom[5] = ts_acf[2]
    avg_vec_mom[6] = ts_acf[3]
    avg_vec_mom[7] = ts_acf[4]
    avg_vec_mom[8] = ts_acf[5]

    # first 4 moments and auto-correlations of the absolute differences of the time series
    abs_diff = np.absolute(np.diff(time_series))

    avg_vec_mom[9] = np.mean(abs_diff)
    avg_vec_mom[10] = np.std(abs_diff)
    s, k = skew(abs_diff), kurtosis(abs_diff)
    avg_vec_mom[11] = np.sign(s) * np.power(abs(s), 1.0 / 3.0)
    avg_vec_mom[12] = np.sign(k) * np.power(abs(k), 1.0 / 4.0)

    ts_diff_acf = sm.tsa.acf(abs_diff, nlags=5, fft=False)
    avg_vec_mom[13] = ts_diff_acf[1]
    avg_vec_mom[14] = ts_diff_acf[2]
    avg_vec_mom[15] = ts_diff_acf[3]
    avg_vec_mom[16] = ts_diff_acf[4]
    avg_vec_mom[17] = ts_diff_acf[5]

    np.nan_to_num(avg_vec_mom, copy=False)

    return avg_vec_mom


def hp_filter(
    time_series: NDArray[np.float64],
    lamb: float = 1600,
) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
    """Apply the HP filter to a time series.

    Args:
        time_series: the one dimensional series to be filtered
        lamb: the Hodrick-Prescott smoothing parameter, a value of 1600 is suggested for quarterly data

    Returns:
        the cycle and the trend of the HP decomposition of the time series

    References:
        Hodrick, R.J, and E. C. Prescott. 1980. "Postwar U.S. Business Cycles: An Empirical Investigation."
        Carnegie Mellon University discussion paper no. 451.
    """
    nobs = len(time_series)
    I = sps.eye(nobs, nobs)  # noqa: E741, N806
    offsets = np.array([0, 1, 2])
    data = np.repeat([[1.0], [-2.0], [1.0]], nobs, axis=1)
    K = sps.dia_matrix((data, offsets), shape=(nobs - 2, nobs))  # noqa: N806

    trend = sps.linalg.spsolve(I + lamb * K.T.dot(K), time_series, use_umfpack=True)
    cycle = time_series - trend

    return cycle, trend


def hp_cycle_lamb1600_filter(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """Get the cycle part of the HP filter to a time series using lamb = 1600.

    Args:
        time_series: the one dimensional series to be filtered

    Returns:
        the cycle part of the HP decomposition of the time series
    """
    return hp_filter(time_series, lamb=1600)[0]


def log_and_hp_filter(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """Take the log of the values and remove the HP trend from them.

    Args:
        time_series: the one dimensional series to be filtered

    Returns:
        the filtered time series
    """
    return np.log(time_series) - hp_filter(np.log(time_series), lamb=1600)[1]


def diff_log_demean_filter(time_series: NDArray[np.float64]) -> NDArray[np.float64]:
    """Take the difference of log of the values and remove the mean from them.

    Args:
        time_series: the one dimensional series to be filtered

    Returns:
        the filtered time series
    """
    log = np.log(time_series)
    diff_log = np.diff(log, prepend=log[0])
    return diff_log - np.mean(diff_log)



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the definition of a 'seedable' base class."""
from __future__ import annotations

from typing import TYPE_CHECKING

from numpy.random import default_rng

if TYPE_CHECKING:
    import numpy as np


class BaseSeedable:
    """BaseSeedable base class.

    This is the base class for all objects that need to keep a random state.

    In particular, the interface provides the following features:
    - it is able to accept a random seed from the client;
    - it allows the client reset the random seed;
    - it provides a getter to the internal random generator;
    - it allows to sample a random seed (in the range [0, 2^32 - 1]).
    """

    def __init__(
        self,
        random_state: int | None = None,
    ) -> None:
        """Initialize the sampler.

        Args:
            random_state: the internal state of the sampler, fixing this numbers the object (e.g. a calibrator,
                a sampler, or a scheduler) behaves deterministically
        """
        self.__random_state: int | None
        self.__random_generator: np.random.Generator

        # this triggers the property setter
        self.random_state = random_state

    @property
    def random_state(self) -> int | None:
        """Get the random state."""
        return self.__random_state

    @random_state.setter
    def random_state(self, random_state: int | None) -> None:
        """Set the random state."""
        self._set_random_state(random_state)

    def _set_random_state(self, random_state: int | None) -> None:
        """Set the random state, private use."""
        self.__random_state = random_state
        self.__random_generator = default_rng(self.random_state)

    @property
    def random_generator(self) -> np.random.Generator:
        """Get the random generator."""
        return self.__random_generator

    def _get_random_seed(self) -> int:
        """Get new random seed from the current random generator."""
        return get_random_seed(self.__random_generator)


def get_random_seed(random_generator: np.random.Generator) -> int:
    """Get a random seed from a random generator.

    Sample an integer in the range [0, 2^32 - 1].

    Args:
        random_generator: the random generator to be used for sampling the random seed.

    Returns:
        the random seed.
    """
    return random_generator.integers(2**32 - 1)



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module defines the 'BaseScheduler' base class."""
from __future__ import annotations

import contextlib
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from black_it.utils.seedable import BaseSeedable

if TYPE_CHECKING:
    from collections.abc import Generator, Sequence

    import numpy as np
    from numpy.typing import NDArray

    from black_it.samplers.base import BaseSampler


class BaseScheduler(BaseSeedable, ABC):
    """BaseScheduler interface.

    This is the base class for all schedulers.
    """

    def __init__(
        self,
        samplers: Sequence[BaseSampler],
        random_state: int | None = None,
    ) -> None:
        """Initialize the scheduler.

        Args:
            samplers: the list of samplers to be scheduled
            random_state: the random seed for the scheduler behaviour
        """
        # need to set __samplers first because _set_random_state requires samplers to be set
        self._samplers = tuple(samplers)
        BaseSeedable.__init__(self, random_state)

    @property
    def samplers(self) -> tuple[BaseSampler, ...]:
        """Get the sequence of samplers."""
        return self._samplers

    def _set_random_state(self, random_state: int | None) -> None:
        """Set the random state (private use)."""
        super()._set_random_state(random_state)
        for sampler in self.samplers:
            sampler.random_state = self._get_random_seed()

    def start_session(self) -> None:
        """Set up the scheduler for a new session.

        The default is a no-op.
        """

    @abstractmethod
    def get_next_sampler(self) -> BaseSampler:
        """Get the sampler to use for the next batch."""

    @abstractmethod
    def update(
        self,
        batch_id: int,
        new_params: NDArray[np.float64],
        new_losses: NDArray[np.float64],
        new_simulated_data: NDArray[np.float64],
    ) -> None:
        """Update the state of the scheduler after each batch.

        Args:
            batch_id: the batch id of the . Must be an integer equal or greater than 0.
            new_params: the new set of parameters sampled in this batch.
            new_losses: the new set of losses corresponding to the batch.
            new_simulated_data: the new set of simulated data, one for each sampled parameter.
        """

    def end_session(self) -> None:
        """Tear down the scheduler at the end of the session.

        The default is a no-op.
        """

    @contextlib.contextmanager
    def session(self) -> Generator[None, None, None]:
        """Start the session of the scheduler with a context manager."""
        self.start_session()
        yield
        self.end_session()



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module defines the 'BaseSampler' base class."""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import numpy as np

from black_it.utils.seedable import BaseSeedable

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from black_it.search_space import SearchSpace


class BaseSampler(BaseSeedable, ABC):
    """BaseSampler interface.

    This is the base class for all samplers.
    """

    def __init__(
        self,
        batch_size: int,
        random_state: int | None = None,
        max_deduplication_passes: int = 5,
    ) -> None:
        """Initialize the sampler.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the internal state of the sampler, fixing this numbers the sampler behaves deterministically
            max_deduplication_passes: maximum number of duplication passes done to avoid sampling repeated parameters
        """
        BaseSeedable.__init__(self, random_state=random_state)
        self.batch_size: int = batch_size
        self.max_deduplication_passes = max_deduplication_passes

    @abstractmethod
    def sample_batch(
        self,
        batch_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Sample a number of new parameters fixed by the 'batch_size' attribute.

        Args:
            batch_size: number of samples to collect
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters

        Returns:
            the new parameters
        """

    def sample(
        self,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],
        existing_losses: NDArray[np.float64],
    ) -> NDArray[np.float64]:
        """Sample from the search space.

        Args:
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled
            existing_losses: the loss corresponding to the sampled parameters

        Returns:
            the sampled parameters
        """
        samples = self.sample_batch(
            self.batch_size,
            search_space,
            existing_points,
            existing_losses,
        )

        for n in range(self.max_deduplication_passes):
            duplicates = self.find_and_get_duplicates(samples, existing_points)

            num_duplicates = len(duplicates)

            if num_duplicates == 0:
                break

            new_samples = self.sample_batch(
                num_duplicates,
                search_space,
                existing_points,
                existing_losses,
            )
            samples[duplicates] = new_samples

            if n == self.max_deduplication_passes - 1:
                print(
                    f"Warning: Repeated samples still found after {self.max_deduplication_passes} duplication passes."
                    " This is probably due to a small search space.",
                )

        return samples

    @staticmethod
    def find_and_get_duplicates(
        new_points: NDArray[np.float64],
        existing_points: NDArray[np.float64],
    ) -> list:
        """Find the points in 'new_points' that are already present in 'existing_points'.

        Args:
            new_points: candidates points for the sampler
            existing_points: previously sampled points

        Returns:
            the location of the duplicates in 'new_points'
        """
        all_points = np.concatenate((existing_points, new_points))
        unq, count = np.unique(all_points, axis=0, return_counts=True)
        repeated_groups = unq[count > 1]

        repeated_pos = []
        if len(repeated_groups) > 0:
            for repeated_group in repeated_groups:
                repeated_idx = np.argwhere(np.all(new_points == repeated_group, axis=1))
                for index in repeated_idx:
                    repeated_pos.append(index[0])  # noqa: PERF401

        return repeated_pos



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains utilities for plotting results."""
from __future__ import annotations

import pickle  # nosec B403
from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt  # type: ignore[import]
import numpy as np
import pandas as pd  # type: ignore[import]
import seaborn as sns  # type: ignore[import]
from ipywidgets import fixed, interact  # type: ignore[import]

from black_it.calibrator import Calibrator

if TYPE_CHECKING:
    import os
    from collections.abc import Collection

    from numpy.typing import NDArray


def _get_samplers_id_table(saving_folder: str | os.PathLike) -> dict[str, int]:
    """Get the id table of the samplers from the checkpoint.

    Args:
        saving_folder: the folder where the calibration results were saved

    Returns:
        the id table of the samplers
    """
    output_file = Path(saving_folder) / "samplers_pickled.pickle"
    with output_file.open("rb") as f:
        method_list = pickle.load(f)  # nosec B301

    return Calibrator._construct_samplers_id_table(method_list)  # noqa: SLF001


def _get_samplers_names(
    saving_folder: str | os.PathLike,
    ids: list[int],
) -> list[str]:
    """Get the names of the samplers from their ids and from the checkpoint of the calibration.

    Args:
        saving_folder: the folder where the calibration results were saved
        ids: the ids of the samplers

    Returns:
        the names of the samplers
    """
    samplers_id_table = _get_samplers_id_table(saving_folder)

    inv_samplers_id_table = {v: k for k, v in samplers_id_table.items()}

    return [inv_samplers_id_table[sampler_id] for sampler_id in ids]


def plot_convergence(saving_folder: str | os.PathLike) -> None:
    """Plot the loss values sampled by the various methods as a function of the batch number.

    Args:
        saving_folder: the folder where the calibration results were saved

    """
    calibration_results_file = Path(saving_folder) / "calibration_results.csv"
    data_frame = pd.read_csv(calibration_results_file)

    losses_cummin = data_frame.groupby("batch_num_samp").min()["losses_samp"].cummin()

    plt.figure()
    g = sns.lineplot(
        data=data_frame,
        x="batch_num_samp",
        y="losses_samp",
        hue="method_samp",
        palette="tab10",
    )

    g = sns.lineplot(
        x=np.arange(max(data_frame["batch_num_samp"]) + 1),
        y=losses_cummin,
        color="black",
        ls="--",
        marker="o",
        label="min loss",
    )

    ids = data_frame["method_samp"].unique()
    sampler_names = _get_samplers_names(saving_folder, ids)

    handles, labels = g.get_legend_handles_labels()
    labels = [*sampler_names, labels[-1]]

    plt.legend(handles, labels, loc="upper right")


def plot_losses(saving_folder: str | os.PathLike) -> None:
    """Plot the parameter sampled colored according to their loss value.

    Args:
        saving_folder: the folder where the calibration results were saved
    """
    calibration_results_file = Path(saving_folder) / "calibration_results.csv"
    data_frame = pd.read_csv(calibration_results_file)

    num_params = sum("params_samp_" in c_str for c_str in data_frame.columns)

    variables = ["params_samp_" + str(i) for i in range(num_params)]

    g = sns.pairplot(
        data_frame,
        hue="losses_samp",
        vars=variables,
        diag_kind="hist",
        corner=True,
        palette="viridis",
        plot_kws={"markers": "o", "linewidth": 1, "alpha": 0.8},
        diag_kws={
            "fill": False,
            "hue": None,
        },
    )

    g.legend.set_bbox_to_anchor((0.8, 0.5))


def plot_sampling(saving_folder: str | os.PathLike) -> None:
    """Plot the parameter sampled colored according to the sampling method used to sample them.

    Args:
        saving_folder: the folder where the calibration results were saved
    """
    calibration_results_file = Path(saving_folder) / "calibration_results.csv"
    data_frame = pd.read_csv(calibration_results_file)

    num_params = sum("params_samp_" in c_str for c_str in data_frame.columns)
    variables = ["params_samp_" + str(i) for i in range(num_params)]

    g = sns.pairplot(
        data_frame,
        hue="method_samp",
        vars=variables,
        diag_kind="hist",
        corner=True,
        palette="tab10",
        plot_kws={"markers": "o", "linewidth": 1, "alpha": 0.8},
        diag_kws={
            "fill": False,
        },
    )
    ids = data_frame["method_samp"].unique()
    sampler_names = _get_samplers_names(saving_folder, ids)

    # take legend of the plot in the last row and first column, to be sure it's a scatter plot
    handles, _ = g.axes[-1][0].get_legend_handles_labels()
    g.legend.remove()

    plt.legend(loc=2, handles=handles, labels=sampler_names, bbox_to_anchor=(0.0, 1.8))


def plot_losses_method_num(
    saving_folder: str | os.PathLike,
    method_num: int,
) -> None:
    """Plot the parameter sampled by a specific sampling method, and color them according to their loss value.

    Args:
        saving_folder: the folder where the calibration results were saved
        method_num: the integer value defining a specific sampling method
    """
    calibration_results_file = Path(saving_folder) / "calibration_results.csv"
    data_frame = pd.read_csv(calibration_results_file)

    if method_num not in set(data_frame["method_samp"]):
        msg = f"Samplers with method_num = {method_num} was never used"
        raise ValueError(msg)

    data_frame = data_frame.loc[data_frame["method_samp"] == method_num]

    num_params = sum("params_samp_" in c_str for c_str in data_frame.columns)

    variables = ["params_samp_" + str(i) for i in range(num_params)]

    g = sns.pairplot(
        data_frame,
        hue="losses_samp",
        vars=variables,
        diag_kind="hist",
        corner=True,
        palette="viridis",
        plot_kws={"markers": "o", "linewidth": 1, "alpha": 0.8},
        diag_kws={
            "fill": False,
            "hue": None,
        },
    )

    g.legend.set_bbox_to_anchor((0.8, 0.5))


def plot_losses_interact(saving_folder: str | os.PathLike) -> None:
    """Plot the parameter sampled colored according to their loss value.

    This plot allows to interactively choose the sampling method.

    Args:
        saving_folder: the folder where the calibration results were saved
    """
    calibration_results_file = Path(saving_folder) / "calibration_results.csv"
    data_frame = pd.read_csv(calibration_results_file)
    method_nums = set(data_frame["method_samp"])

    samplers_id_table = _get_samplers_id_table(saving_folder)

    id_samplers_cal = {}

    for sampler, sampler_id in samplers_id_table.items():
        if sampler_id in method_nums:
            id_samplers_cal[sampler] = sampler_id  # noqa: PERF403

    interact(
        plot_losses_method_num,
        saving_folder=fixed(saving_folder),
        method_num=id_samplers_cal,
    )


def plot_sampling_batch_nums(
    saving_folder: str | os.PathLike,
    batch_nums: Collection[int],
) -> None:
    """Plot the parameter sampled in specific batches colored according to the sampling method used to sample them.

    Args:
        saving_folder: the folder where the calibration results were saved
        batch_nums: a list of batch number
    """
    calibration_results_file = Path(saving_folder) / "calibration_results.csv"
    data_frame = pd.read_csv(calibration_results_file)

    # deduplicate batch numbers
    batch_nums = set(batch_nums)

    filter_bns = [bn in batch_nums for bn in data_frame["batch_num_samp"]]

    data_frame["filter_bns"] = filter_bns

    data_frame_2 = data_frame.loc[data_frame["filter_bns"] == True]  # noqa: E712

    num_params = sum("params_samp_" in c_str for c_str in data_frame.columns)

    variables = ["params_samp_" + str(i) for i in range(num_params)]

    g = sns.pairplot(
        data_frame_2,
        hue="method_samp",
        vars=variables,
        diag_kind="hist",
        corner=True,
        plot_kws={"markers": "o", "linewidth": 1, "alpha": 0.8},
        diag_kws={"fill": False},
        palette="tab10",
    )

    ids = data_frame["method_samp"].unique()
    sampler_names = _get_samplers_names(saving_folder, ids)

    # take legend of the plot in the last row and first column, to be sure it's a scatter plot
    handles, _ = g.axes[-1][0].get_legend_handles_labels()
    g.legend.remove()

    plt.legend(loc=2, handles=handles, labels=sampler_names, bbox_to_anchor=(0.0, 1.8))


def plot_sampling_interact(saving_folder: str | os.PathLike) -> None:
    """Plot the parameter sampled colored according to the sampling method used to sample them.

    The method allows to interactively choose the batch numbers included in the plot.

    Args:
        saving_folder: the folder where the calibration results were saved
    """
    calibration_results_file = Path(saving_folder) / "calibration_results.csv"
    data_frame = pd.read_csv(calibration_results_file)

    max_bn = int(max(data_frame["batch_num_samp"]))
    all_bns: NDArray[np.int64] = np.arange(max_bn + 1, dtype=int)
    indices_bns = np.array_split(all_bns, min(max_bn, 3))

    dict_bns = {}
    for index_bns in indices_bns:
        key = "from " + str(index_bns[0]) + " to " + str(index_bns[-1])
        dict_bns[key] = index_bns.tolist()

    interact(
        plot_sampling_batch_nums,
        saving_folder=fixed(saving_folder),
        batch_nums=dict_bns,
    )



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the quadratic loss."""
from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from scipy.spatial.distance import minkowski  # type: ignore[import]

from black_it.loss_functions.base import BaseLoss

if TYPE_CHECKING:
    import numpy as np
    from numpy.typing import NDArray


class MinkowskiLoss(BaseLoss):
    """Class for the Minkowski loss."""

    def __init__(
        self,
        p: int = 2,
        coordinate_weights: NDArray | None = None,
        coordinate_filters: list[Callable | None] | None = None,  # noqa: ARG002
    ) -> None:
        """Loss computed using a Minkowski distance.

        The [Minkowski distance](https://en.wikipedia.org/wiki/Minkowski_distance)
        is a generalization of both the Manhattan distance (p=1) and the Euclidean distance (p=2).

        This function computes the Minkowski distance between two series.

        Note: this class is a wrapper of scipy.spatial.distance.minkowski

        Args:
            p: The order of the norm used to compute the distance between real and simulated series
            coordinate_weights: The order of the norm used to compute the distance between real and simulated series
            coordinate_filters: filters/transformations to be applied to each simulated series before
                the loss computation.
        """
        self.p = p
        super().__init__(coordinate_weights)

    def compute_loss_1d(
        self,
        sim_data_ensemble: NDArray[np.float64],
        real_data: NDArray[np.float64],
    ) -> float:
        """Call scipy.spatial.distance.minkowski() on its arguments.

        Args:
            sim_data_ensemble: the first operand
            real_data: the second operand

        Returns:
            The computed loss over the specific coordinate.
        """
        # average simulated time series
        sim_data_ensemble = sim_data_ensemble.mean(axis=0)

        return minkowski(sim_data_ensemble, real_data, p=self.p)



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the main class Calibrator."""
from __future__ import annotations

import multiprocessing
import textwrap
import time
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Callable, cast

import numpy as np
from joblib import Parallel, delayed  # type: ignore[import]

from black_it.schedulers.base import BaseScheduler
from black_it.schedulers.round_robin import RoundRobinScheduler
from black_it.search_space import SearchSpace
from black_it.utils.base import _assert
from black_it.utils.json_pandas_checkpointing import (
    load_calibrator_state,
    save_calibrator_state,
)
from black_it.utils.seedable import BaseSeedable

if TYPE_CHECKING:
    import os
    from collections.abc import Sequence

    from numpy.typing import NDArray

    from black_it.loss_functions.base import BaseLoss
    from black_it.samplers.base import BaseSampler


class Calibrator(BaseSeedable):
    """The class used to perform a calibration."""

    STATE_VERSION = 0

    def __init__(  # noqa: PLR0913
        self,
        loss_function: BaseLoss,
        real_data: NDArray[np.float64],
        model: Callable,
        parameters_bounds: NDArray[np.float64] | list[list[float]],
        parameters_precision: NDArray[np.float64] | list[float],
        ensemble_size: int,
        samplers: Sequence[BaseSampler] | None = None,
        scheduler: BaseScheduler | None = None,
        sim_length: int | None = None,
        convergence_precision: int | None = None,
        verbose: bool = True,
        saving_folder: str | None = None,
        random_state: int | None = None,
        n_jobs: int | None = None,
    ) -> None:
        """Initialize the Calibrator object.

        It must be initialized with details on the parameters to explore,
        on the model to calibrate, on the samplers and on the loss function to use.

        Args:
            loss_function: a loss function which evaluates the similarity between simulated and real datasets
            real_data: an array containing the real time series
            model: a model with free parameters to be calibrated
            parameters_bounds: the bounds of the parameter space
            parameters_precision: the precisions to be used for the discretization of the parameters
            ensemble_size: number of repetitions to be run for each set of parameters to decrease statistical
                fluctuations. For deterministic models this should be set to 1.
            samplers: list of methods to be used in the calibration procedure
            scheduler: the scheduler to be used in order to schedule the available samplers
            sim_length: number of periods to simulate the model for, by default this is equal to the length of the
                real time series.
            convergence_precision: number of significant digits to consider in the convergence check. The check is
                not performed if this is set to 'None'.
            verbose: whether to print calibration updates
            saving_folder: the name of the folder where data should be saved and/or retrieved
            random_state: random state of the calibrator, used for model simulations and to initialise the samplers
            n_jobs: the maximum number of concurrently running jobs. For more details, see the
                [joblib.Parallel documentation](https://joblib.readthedocs.io/en/latest/generated/joblib.Parallel.html).

        """
        BaseSeedable.__init__(self, random_state=random_state)
        self.loss_function = loss_function
        self.model = model
        self.random_state = random_state
        self.real_data = real_data
        self.ensemble_size = ensemble_size

        if sim_length is None:
            self.N = self.real_data.shape[0]
        else:
            if sim_length != self.real_data.shape[0]:
                warnings.warn(  # noqa: B028
                    "The length of real time series is different from the simulation length, "
                    f"got {self.real_data.shape[0]} and {sim_length}. This may or may not be a problem depending "
                    "on the loss function used.",
                    RuntimeWarning,
                )
            self.N = sim_length
        self.D = self.real_data.shape[1]
        self.verbose = verbose
        self.convergence_precision = (
            self._validate_convergence_precision(convergence_precision)
            if convergence_precision is not None
            else None
        )
        self.saving_folder = saving_folder

        # Initialize search grid
        self.param_grid = SearchSpace(parameters_bounds, parameters_precision, verbose)

        # initialize arrays
        self.params_samp = np.zeros((0, self.param_grid.dims))
        self.losses_samp = np.zeros(0)
        self.batch_num_samp: NDArray[np.int64] = np.zeros(0, dtype=int)
        self.method_samp: NDArray[np.int64] = np.zeros(0, dtype=int)
        self.series_samp: NDArray[np.float64] = np.zeros(
            (0, self.ensemble_size, self.N, self.D),
        )

        # initialize variables before calibration
        self.n_sampled_params = 0
        self.current_batch_index = 0

        # set number of processes for parallel evaluation of model
        self.n_jobs = n_jobs if n_jobs is not None else multiprocessing.cpu_count()

        print(
            f"Selecting {self.n_jobs} processes for the parallel evaluation of the model",
        )

        self.scheduler = self.__validate_samplers_and_scheduler_constructor_args(
            samplers,
            scheduler,
        )

        self.samplers_id_table = self._construct_samplers_id_table(
            list(self.scheduler.samplers),
        )

    @classmethod
    def __validate_samplers_and_scheduler_constructor_args(
        cls,
        samplers: Sequence[BaseSampler] | None,
        scheduler: BaseScheduler | None,
    ) -> BaseScheduler:
        """Validate the 'samplers' and the 'scheduler' arguments provided to the constructor."""
        both_none = samplers is None and scheduler is None
        both_not_none = samplers is not None and scheduler is not None
        if both_none and both_not_none:
            msg = "only one between 'samplers' and 'scheduler' must be provided"
            raise ValueError(
                msg,
            )

        if samplers is not None:
            return RoundRobinScheduler(samplers)

        return cast(BaseScheduler, scheduler)

    def _set_samplers_seeds(self) -> None:
        """Set the calibration seed."""
        self.scheduler.random_state = self.random_state

        # "burn" seeds from the calibrator seed generator for backward compatibility
        for _ in self.scheduler.samplers:
            self._get_random_seed()

    @staticmethod
    def _construct_samplers_id_table(samplers: list[BaseSampler]) -> dict[str, int]:
        """Construct the samplers-by-id table.

        Given the list (built-in or user-defined) of samplers a calibration
        session is going to use, return a map from the sampler human-readable
        name to a numeric id (starting from 0).

        Different calibration sessions may result in different conversion
        tables.

        Args:
            samplers: the list of samplers of the calibrator

        Returns:
            A dict that maps from the given sampler names to unique ids.
        """
        samplers_id_table = {}
        sampler_id = 0

        for sampler in samplers:
            sampler_name = type(sampler).__name__
            if sampler_name in samplers_id_table:
                continue

            samplers_id_table[sampler_name] = sampler_id
            sampler_id = sampler_id + 1

        return samplers_id_table

    def set_samplers(self, samplers: list[BaseSampler]) -> None:
        """Set the samplers list of the calibrator.

        This method overwrites the samplers of the calibrator object with a custom list of samplers.

        Args:
            samplers: a list of samplers

        """
        # overwrite the list of samplers
        self.scheduler._samplers = tuple(samplers)  # noqa: SLF001
        self.update_samplers_id_table(samplers)

    def set_scheduler(self, scheduler: BaseScheduler) -> None:
        """Overwrite the scheduler of the calibrator object."""
        self.scheduler = scheduler
        self.update_samplers_id_table(self.scheduler.samplers)

    def update_samplers_id_table(self, samplers: Sequence[BaseSampler]) -> None:
        """Update the samplers_id_table attribute with possible new samplers."""
        # update the samplers_id_table with the new samplers, only if necessary
        sampler_id = max(self.samplers_id_table.values()) + 1

        for sampler in samplers:
            sampler_name = type(sampler).__name__
            if sampler_name in self.samplers_id_table:
                continue

            self.samplers_id_table[sampler_name] = sampler_id
            sampler_id = sampler_id + 1

    @classmethod
    def restore_from_checkpoint(
        cls,
        checkpoint_path: str,
        model: Callable,
    ) -> Calibrator:
        """Return an instantiated class from a database file and a model simulator.

        Args:
            checkpoint_path: the name of the database file to read from
            model: the model to calibrate. It must be equal to the one already calibrated

        Returns:
            An initialised Calibrator object.
        """
        (
            parameters_bounds,
            parameters_precision,
            real_data,
            ensemble_size,
            sim_length,
            _d,
            convergence_precision,
            verbose,
            saving_file,
            random_state,
            random_generator_state,
            model_name,
            scheduler,
            loss_function,
            current_batch_index,
            n_sampled_params,
            n_jobs,
            params_samp,
            losses_samp,
            series_samp,
            batch_num_samp,
            method_samp,
        ) = load_calibrator_state(checkpoint_path, cls.STATE_VERSION)

        _assert(
            model_name == model.__name__,
            (
                "Error: the model provided appears to be different from the one present "
                "in the database"
            ),
        )

        calibrator = cls(
            loss_function,
            real_data,
            model,
            parameters_bounds,
            parameters_precision,
            ensemble_size,
            scheduler=scheduler,
            sim_length=sim_length,
            convergence_precision=convergence_precision,
            verbose=verbose,
            saving_folder=saving_file,
            random_state=random_state,
            n_jobs=n_jobs,
        )

        calibrator.current_batch_index = current_batch_index
        calibrator.n_sampled_params = n_sampled_params
        calibrator.params_samp = params_samp
        calibrator.losses_samp = losses_samp
        calibrator.series_samp = series_samp
        calibrator.batch_num_samp = batch_num_samp
        calibrator.method_samp = method_samp

        # reset the random number generator state
        calibrator.random_generator.bit_generator.state = random_generator_state

        return calibrator

    def simulate_model(self, params: NDArray) -> NDArray:
        """Simulate the model.

        This method calls the model simulator in parallel on a given set of parameter values, a number of repeated
        evaluations are performed for each parameter to average out random fluctuations.

        Args:
            params: the array of parameters for which the model should be evaluated
        # noqa
        Returns:
            simulated_data: an array of dimensions (batch_size, ensemble_size, N, D) containing all
                simulated time series
        """
        rep_params = np.repeat(params, self.ensemble_size, axis=0)

        simulated_data_list = Parallel(n_jobs=self.n_jobs)(
            delayed(self.model)(param, self.N, self._get_random_seed())
            for i, param in enumerate(rep_params)
        )

        simulated_data = np.array(simulated_data_list)

        return np.reshape(
            simulated_data,
            (params.shape[0], self.ensemble_size, self.N, self.D),
        )

    def calibrate(self, n_batches: int) -> tuple[NDArray, NDArray]:
        """Run calibration for n batches.

        Args:
            n_batches (int): number of 'batches' to be executed. Each batch runs over all methods

        Returns:
            The sampled parameters and the corresponding sampled losses.
            Both arrays are sorted by increasing loss values
        """
        if self.current_batch_index == 0:
            # we only set the samplers' random state at the start of a calibration
            self._set_samplers_seeds()

        with self.scheduler.session():
            for _ in range(n_batches):
                print()
                print(f"BATCH NUMBER:   {self.current_batch_index + 1}")
                print(f"PARAMS SAMPLED: {self.n_sampled_params}")

                method = self.scheduler.get_next_sampler()

                t_start = time.time()
                print()
                print(f"METHOD: {type(method).__name__}")

                # get new params from a specific sampler
                new_params = method.sample(
                    self.param_grid,
                    self.params_samp,
                    self.losses_samp,
                )

                t_eval = time.time()

                # simulate an ensemble of models for different parameters

                new_simulated_data = self.simulate_model(new_params)

                new_losses = []

                for sim_data_ensemble in new_simulated_data:
                    new_loss = self.loss_function.compute_loss(
                        sim_data_ensemble,
                        self.real_data,
                    )
                    new_losses.append(new_loss)

                # update arrays
                self.params_samp = np.vstack((self.params_samp, new_params))
                self.losses_samp = np.hstack((self.losses_samp, new_losses))
                self.series_samp = np.vstack((self.series_samp, new_simulated_data))
                self.batch_num_samp = np.hstack(
                    (
                        self.batch_num_samp,
                        [self.current_batch_index] * method.batch_size,
                    ),
                )
                self.method_samp = np.hstack(
                    (
                        self.method_samp,
                        [self.samplers_id_table[type(method).__name__]]
                        * method.batch_size,
                    ),
                )

                # logging
                t_end = time.time()
                if self.verbose:
                    min_dist_new_points = np.round(np.min(new_losses), 2)
                    avg_dist_new_points = np.round(np.average(new_losses), 2)
                    avg_dist_existing_points = np.round(np.average(self.losses_samp), 2)

                    elapsed_tot = np.round(t_end - t_start, 1)
                    elapsed_eval = np.round(t_end - t_eval, 1)
                    print(
                        textwrap.dedent(
                            f"""\
                        ----> sim exec elapsed time: {elapsed_eval}s
                        ---->   min loss new params: {min_dist_new_points}
                        ---->   avg loss new params: {avg_dist_new_points}
                        ----> avg loss exist params: {avg_dist_existing_points}
                        ---->         curr min loss: {np.min(self.losses_samp)}
                        ====>    total elapsed time: {elapsed_tot}s
                        """,
                        ),
                        end="",
                    )

                # update count of number of params sampled
                self.n_sampled_params = self.n_sampled_params + len(new_params)

                self.scheduler.update(
                    self.current_batch_index,
                    new_params,
                    new_losses,  # type: ignore[arg-type]
                    new_simulated_data,
                )

                self.current_batch_index += 1

                # check convergence for early termination
                if self.convergence_precision is not None:
                    converged = self.check_convergence(
                        self.losses_samp,
                        self.n_sampled_params,
                        self.convergence_precision,
                    )
                    if converged and self.verbose:
                        print("\nCONVERGENCE CHECK:")
                        print("Achieved convergence loss, stopping search.")
                        break

                if self.saving_folder is not None:
                    self.create_checkpoint(self.saving_folder)

            idx = np.argsort(self.losses_samp)

        return self.params_samp[idx], self.losses_samp[idx]

    @staticmethod
    def check_convergence(
        losses_samp: NDArray,
        n_sampled_params: int,
        convergence_precision: int,
    ) -> bool:
        """Check convergence of the calibration.

        Args:
            losses_samp: the sampled losses
            n_sampled_params: the number of sampled params
            convergence_precision: the required convergence precision.

        Returns:
            True if the calibration converged, False otherwise.
        """
        return (
            np.round(np.min(losses_samp[:n_sampled_params]), convergence_precision) == 0
        )

    def create_checkpoint(self, file_name: str | os.PathLike) -> None:
        """Save the current state of the object.

        Args:
            file_name: the name of the folder where the data will be saved
        """
        checkpoint_path: Path = Path(file_name).resolve()

        t_start = time.time()

        model_name = self.model.__name__

        save_calibrator_state(
            checkpoint_path,
            self.param_grid.parameters_bounds,
            self.param_grid.parameters_precision,
            self.real_data,
            self.ensemble_size,
            self.N,
            self.D,
            self.convergence_precision,
            self.verbose,
            self.saving_folder,
            self.random_state,
            self.random_generator.bit_generator.state,
            model_name,
            self.scheduler,
            self.loss_function,
            self.current_batch_index,
            self.n_sampled_params,
            self.n_jobs,
            self.params_samp,
            self.losses_samp,
            self.series_samp,
            self.batch_num_samp,
            self.method_samp,
        )

        t_end = time.time()

        elapsed = np.round(t_end - t_start, 1)
        print(f"Checkpoint saved in {elapsed}s")

    @staticmethod
    def _validate_convergence_precision(convergence_precision: int) -> int:
        """Validate convergence precision input."""
        _assert(
            convergence_precision >= 0,
            f"convergence precision must be an integer greater than 0, got {convergence_precision}",
            exception_class=ValueError,
        )
        return convergence_precision



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module defines the 'LikelihoodLoss'  class."""
from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Callable, cast

import numpy as np

from black_it.loss_functions.base import BaseLoss

if TYPE_CHECKING:
    from numpy.typing import NDArray


def kernel(sq_dist: NDArray[np.float64], h: float, d: int) -> NDArray[np.float64]:
    """Compute a kernel density estimation using a Gaussian kernel."""
    return np.exp(-(sq_dist / (2 * h**2))) / (h**d * (2 * np.pi) ** (d / 2.0))


class LikelihoodLoss(BaseLoss):
    """Class for a likelihood loss computation using a Gaussian kernel function.

    The likelihood computation can be useful within the approach proposed in
    "Grazzini et al, Bayesian estimation of agent-based models, JEDC (2017)".

    """

    def __init__(
        self,
        coordinate_weights: NDArray | None = None,
        coordinate_filters: list[Callable | None] | None = None,
        h: str | float = "silverman",
    ) -> None:
        """Initialize the loss function.

        Args:
            coordinate_weights: the weights of the loss coordinates.
            coordinate_filters: filters/transformations to be applied to each simulated series before
                the loss computation.
            h: the bandwidth of the kernel used to estimate the likelihood
        """
        super().__init__(coordinate_weights, coordinate_filters)
        self.h = h

    @staticmethod
    def _get_bandwidth_silverman(n: int, d: int) -> float:
        """Return a reasonable bandwidth value computed using the Silverman's rule of thumb."""
        return ((n * (d + 2)) / 4) ** (-1 / (d + 4))

    @staticmethod
    def _get_bandwidth_scott(n: int, d: int) -> float:
        """Return a reasonable bandwidth value computed using the Scott's rule of thumb."""
        return n ** (-1 / (d + 4))

    def compute_loss(
        self,
        sim_data_ensemble: NDArray[np.float64],
        real_data: NDArray[np.float64],
    ) -> float:
        """Compute the loss between simulated and real data.

        Args:
            sim_data_ensemble: an ensemble of simulated data, of shape (ensemble_size, N, D)
            real_data: the real data, of shape (T, D)

        Returns:
            The loss value.
        """
        sim_data_ensemble_shape: tuple[int, int, int] = cast(
            tuple[int, int, int],
            sim_data_ensemble.shape,
        )
        r = sim_data_ensemble_shape[0]  # number of repetitions
        s = sim_data_ensemble_shape[1]  # simulation length
        d = sim_data_ensemble_shape[2]  # time series dimension

        if self.coordinate_weights is not None:
            warnings.warn(  # noqa: B028
                "The LikelihoodLoss cannot take any coordinate weights. "
                "The provided coordinate_weights will be ignored",
                RuntimeWarning,
            )

        filters = self._check_coordinate_filters(d)
        filtered_data = self._filter_data(filters, sim_data_ensemble)
        sim_data_ensemble = np.transpose(filtered_data, (1, 2, 0))

        h = self._check_bandwidth(s, d)
        sq_dists_r_t_s = (
            1.0
            / d
            * np.sum(
                (sim_data_ensemble[:, None, :, :] - real_data[None, :, None, :]) ** 2,
                axis=3,
            )
        )

        kernel_r_t_s = kernel(sq_dists_r_t_s, h, d)
        lik_real_series_r_t = np.sum(kernel_r_t_s, axis=2) / s
        log_lik_real_series_r = np.sum(np.log(lik_real_series_r_t), axis=1)
        log_lik_real_series = np.sum(log_lik_real_series_r, axis=0) / r
        return -log_lik_real_series

    def _check_bandwidth(self, s: int, d: int) -> float:
        """Check the bandwidth self.h and return a usable one."""
        h: float

        if isinstance(self.h, str):
            if self.h == "silverman":
                h = self._get_bandwidth_silverman(s, d)
            elif self.h == "scott":
                h = self._get_bandwidth_scott(s, d)
            else:
                msg = (
                    "Select a valid rule of thumb (either 'silverman' or 'scott') "
                    "or directly a numerical value for the bandwidth"
                )
                raise KeyError(
                    msg,
                )
        else:
            h = self.h

        return h

    def compute_loss_1d(
        self,  # noqa: PLR6301
        sim_data_ensemble: NDArray[np.float64],  # noqa: ARG002
        real_data: NDArray[np.float64],  # noqa: ARG002
    ) -> float:
        """Compute likelihood loss on a single dimension, not available."""
        msg = "The likelihood cannot be currently computed on a single dimension."
        raise NotImplementedError(
            msg,
        )



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This module contains the implementation of the R-sequence sampler."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

from black_it.samplers.base import BaseSampler
from black_it.utils.base import check_arg, digitize_data

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from black_it.search_space import SearchSpace

_MIN_SEQUENCE_START_INDEX = 20
_MAX_SEQUENCE_START_INDEX = 2**16


class RSequenceSampler(BaseSampler):
    """The R-sequence sampler."""

    def __init__(
        self,
        batch_size: int,
        random_state: int | None = None,
        max_deduplication_passes: int = 5,
    ) -> None:
        """Initialize the sampler.

        Args:
            batch_size: the number of points sampled every time the sampler is called
            random_state: the random state of the sampler, fixing this number the sampler behaves deterministically
            max_deduplication_passes: (non-negative integer) the maximum number of deduplication passes that are made
                after every batch sampling. Default: 0, i.e. no deduplication happens.
        """
        super().__init__(batch_size, random_state, max_deduplication_passes)

        self._sequence_index: int
        self._sequence_start: float
        self._reset()

    @classmethod
    def compute_phi(cls, nb_dims: int) -> float:
        """Get an approximation of phi^nb_dims.

        Args:
            nb_dims: the number of dimensions.

        Returns:
            phi^nb_dims
        """
        check_arg(nb_dims >= 1, f"nb_dims should be greater than 0, got {nb_dims}")
        phi: float = 2.0
        old_phi = None
        while old_phi != phi:
            old_phi = phi
            phi = pow(1 + phi, 1.0 / (nb_dims + 1))
        return phi

    def _set_random_state(self, random_state: int | None) -> None:
        """Set the random state (private use).

        For the RSequence sampler, it also resets the sequence index and the sequence start.

        Args:
            random_state: the random seed
        """
        super()._set_random_state(random_state)
        self._reset()

    def _reset(self) -> None:
        """Reset the index of the sequence."""
        self._sequence_index = self.random_generator.integers(
            _MIN_SEQUENCE_START_INDEX,
            _MAX_SEQUENCE_START_INDEX,
        )
        self._sequence_start = self.random_generator.random()

    def sample_batch(
        self,
        batch_size: int,
        search_space: SearchSpace,
        existing_points: NDArray[np.float64],  # noqa: ARG002
        existing_losses: NDArray[np.float64],  # noqa: ARG002
    ) -> NDArray[np.float64]:
        """Sample points using the R-sequence.

        Args:
            batch_size: the number of samples
            search_space: an object containing the details of the parameter search space
            existing_points: the parameters already sampled (not used)
            existing_losses: the loss corresponding to the sampled parameters (not used)

        Returns:
            the parameter sampled
        """
        unit_cube_points: NDArray[np.float64] = self._r_sequence(
            batch_size,
            search_space.dims,
        )
        p_bounds: NDArray[np.float64] = search_space.parameters_bounds
        sampled_points = p_bounds[0] + unit_cube_points * (p_bounds[1] - p_bounds[0])
        return digitize_data(sampled_points, search_space.param_grid)

    def _r_sequence(self, nb_samples: int, dims: int) -> NDArray[np.float64]:
        """Compute the R-sequence (http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/).

        Args:
            nb_samples: number of points to sample
            dims: the number of dimensions

        Returns:
            Set of params uniformly placed in d-dimensional unit cube.
        """
        phi = self.compute_phi(dims)
        alpha: NDArray[np.float64] = np.power(1 / phi, np.arange(1, dims + 1)).reshape(
            (1, -1),
        )
        end_index = self._sequence_index + nb_samples
        indexes = np.arange(self._sequence_index, end_index).reshape((-1, 1))
        points: NDArray[np.float64] = (self._sequence_start + indexes.dot(alpha)) % 1
        self._sequence_index = end_index
        return points



# Black-box ABM Calibration Kit (Black-it)
# Copyright (C) 2021-2024 Banca d'Italia
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""This package contains implementations of Gym environments for the RL Scheduler."""
from __future__ import annotations

from abc import ABC, abstractmethod
from queue import Queue
from typing import TYPE_CHECKING, Any, SupportsFloat

import gymnasium as gym
import numpy as np
from gymnasium.core import ObsType, RenderFrame
from gymnasium.spaces import Discrete

from black_it.utils.base import _assert

if TYPE_CHECKING:
    from numpy._typing import NDArray


class CalibrationEnv(gym.Env[ObsType, np.int64], ABC):
    """A Gym environment that wraps the calibration task."""

    def __init__(self, nb_samplers: int) -> None:
        """Initialize the environment."""
        self._nb_samplers = nb_samplers
        self._out_queue: Queue = Queue()
        self._in_queue: Queue = Queue()

        self._curr_best_loss: float | None = None

        self.action_space = Discrete(self._nb_samplers)

    @abstractmethod
    def reset_state(self) -> ObsType:
        """Get the initial state."""

    @abstractmethod
    def get_reward(self, best_param: NDArray, best_loss: float) -> float:
        """Get the current reward."""

    @abstractmethod
    def get_next_observation(self) -> ObsType:
        """Get the next observation."""

    def reset(
        self,
        *,
        seed: int | None = None,  # noqa: ARG002
        options: dict[str, Any] | None = None,  # noqa: ARG002
    ) -> tuple[ObsType, dict[str, Any]]:
        """Reset the environment."""
        return self.reset_state(), {}

    def step(
        self,
        action: np.int64,
    ) -> tuple[ObsType, SupportsFloat, bool, bool, dict[str, Any]]:
        """Do a step."""
        _assert(self.action_space.contains(action))
        self._out_queue.put(action)
        result = self._in_queue.get()
        if result is None:
            # calibration ended
            return self.reset_state(), 0.0, False, True, {}
        best_param, best_loss = result
        reward = self.get_reward(best_param, best_loss)
        next_obs = self.get_next_observation()
        return next_obs, reward, False, False, {}

    def render(self) -> RenderFrame | list[RenderFrame] | None:
        """Render the environment (not implemented)."""
        raise NotImplementedError
